'''
Created on Jan 17, 2021

@author: vladislavkargin
'''

#import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm

class Ribbon(object):
    '''
    This class realizes a ribbon tile for ribbon tilings.
    
    Ribbon is deetermined by its root square 
    s_{xy} (south-west corner is at (x, y)) and its shape:  
    a sequence of (n-1) zeros and ones. 
    0 means go right, 1 means go up. This list can be empty which will 
    corresponds to a ribbon tile that consist of a single square. 
    
    The ribbon has also attribute squares
    which is the list of squares in the ribbon.   
    '''


    def __init__(self, x, y, shape):
        '''
        creates a ribbon with root square s_{xy} (south-west corner is at (x, y))
        and a list of 0 and 1 that determines the shape.   
        Also creates a list of squares inside the ribbon. 
        '''
        self.x = x 
        self.y = y 
        self.shape = shape
        s = x
        t = y
        self.squares = [(s, t)]
        for d in shape:
            if d == 0:
                s = s + 1
                self.squares.append((s, t))
            else:
                t = t + 1
                self.squares.append((s, t))
     
    def __str__(self):
        '''
        Returns a string representation of this ribbon.
        return the graph
        '''
        s = '(' + str(self.x) + ',' + str(self.y) + '), ' + str(self.shape)
        return s   
                
    def level(self):
        ''' return the level of the ribbon '''
        return self.x + self.y
         
    def contains(self, x, y):
        ''' check if square (x, y) belong to ribbon 
        
        Parameters
        ----------
        x, y - integers
            The coordinates of the square
        Returns
        ----------
        boolean 
            true if the square is in the ribbon
        '''
        return (x, y) in self.squares
    
    def flip(self, B):
        ''' check if this ribbon can be flipped with ribbon B. If yes, returns (True, A1', B1')
        where A1', B1' is a flipped pair, if no, returns (False, A, B)
        In this version, ribbons are required to be of the same length, 
        although in principle, they can be define for ribbons of different 
        length as well
        
        Parameters
        -----------
        B: ribbon 
           required to  have the same length
        
        Returns
        ----------- 
        returns (True, A1', B1') where A1', B1' is a flipped pair, if the flip is possible;
        (False, A, B) if not possible
        
        '''
    
        n = len(self.shape) #Potential for confusion: n is NOT the size of the ribbon, it is 
                            #size of the shape. So for 3-ribbons, n = 2
        if len(B.shape)!= n:
            return (False, self, B)
        
        lA = self.x + self.y 
        lB = B.x + B.y
        if lA == lB:
            return (False, self, B) #the tiles at the same level can't be flipped
        
        if lA < lB: #we define X and Y so that X has the lower level than Y
            X = Ribbon(self.x, self.y, self.shape) 
            Y = Ribbon(B.x, B.y, B.shape)
        else:
            Y = Ribbon(self.x, self.y, self.shape) 
            X = Ribbon(B.x, B.y, B.shape)
        
        for i in range(n): #we are going along the X and checking what happens if we 
                        #change direction at certain step        
            #print("i = ", i)       
            s, t = X.squares[i]
            flag = True #if remains true after following checks then we can return a flip 
            if X.shape[i] == 0: #instead of going right, go up
                #print("First case")
                t = t + 1
                if Y.squares[0] != (s, t):
                    #print("Breaking because Y[0] != ", s, ",", t)
                    continue
                for j in range(n - i - 1): #for 3-ribbon (n = 2) and i = 0, it is only j = 0
                    #print("j = ", j)
                    if X.shape[i + j + 1] == 0:
                        s = s + 1
                    else: 
                        t = t + 1
                    #print((s,t))
                    if Y.squares[j + 1] != (s, t):
                        #print("Breaking because Y[j+1] != ", s, ",", t)
                        flag = False
                        break
                if flag and Y.shape[n - i - 1] == 0:
                    #print(i)
                    A1 = Ribbon(X.x, X.y, X.shape[:i] + [1] + X.shape[i+1:])
                    B1 = Ribbon(X.squares[i + 1][0], X.squares[i+1][1], Y.shape[:(n-i -1)] + [1] + Y.shape[(n-i):])
                    #print("first case of flippability satisfied at i = ", i)
                    #we want to make sure that the 
                    #the ribbons are returned in the correct order
                    #(without switching levels)
                    if A1.x + A1.y == self.x + self.y:
                        return (True, A1, B1)
                    else: 
                        return (True, B1, A1)
            else: #X.shape is 1, instead of going up, go right
                #print("Second case")
                s = s + 1
                if Y.squares[0] != (s, t):
                    #print("Breaking because Y[0] != ", s, ",", t)
                    continue
                for j in range(n - i - 1): #for certain amount of time, the new ribbon should go
                                           #along the squares of Y and then go up
                    #print("j = ", j)
                    if X.shape[i + j + 1] == 0:
                        s = s + 1
                    else: 
                        t = t + 1
                    #print((s,t))
                    if Y.squares[j + 1] != (s, t):
                        #print("Breaking because Y[j+1] != ", s, ",", t)
                        flag = False
                        break
                if flag and Y.shape[n - i - 1] == 1:
                    #print(i)
                    A1 = Ribbon(X.x, X.y, X.shape[:i] + [0] + X.shape[i+1:])
                    B1 = Ribbon(X.squares[i + 1][0], X.squares[i+1][1], Y.shape[:(n-i -1)] + [0] + Y.shape[(n-i):])
                    #print("second case of flippability satsified for i = ", i)
                    if A1.x + A1.y == self.x + self.y:
                        return (True, A1, B1)
                    else:
                        return (True, B1, A1)
                
        return (False, self, B)
    
    def draw(self, ax = None, 
            colormap = 'prism',
            colorType = "Shape", 
            block = False, MaxLength = None):
        '''
        draw the ribbon.
        
        Parameters:   
        ax : matplotlib axis
            axis in which to draw (default = None)
        colormap : 'prism' 
            the name of the coloring scheme ("palette"), used to color the tiling,
            default = "prism"
            other possibilities: "jet"and many others, see matplotlib docs.
        colorType : string
            defines the type of coloring. Possible choices:
            "Shape" -- color only depends on shape
            "ShapePosition" -- color depends on shape and level (mod n)
            "Length" -- color only depends on length of the ribbon
            (default = "Shape")
            
        MaxLength : integer 
            a fine-tuning parameter used in the choice of coloring (default = None)
        block: boolean
            if True, stops the execution and draws the graph (default = False)
        
        '''
        if (ax == None):
            fig, ax = plt.subplots()
            ax.set_xlim(0, self.x + len(self.shape) + 1)
            ax.set_ylim(0, self.y + len(self.shape) + 1)
        else:
            fig = ax.figure
        
        l = len(self.shape) # l = n - 1
        cmap = cm.get_cmap(colormap)
        if MaxLength == None:
            MaxLength = l + 2**l #this is actually not maxLength but 
                                #the max parameter for the number of 
                                #colors
        else: 
            MaxLength = max([MaxLength, l])
        
        if colorType == "Length":
            c = cmap(l/MaxLength) #for floats cmap converts interval [0,1]
                                #to the RGB range, so l should be comparable to 
                                # Maxlength
        elif colorType == "ShapePosition":
            #calculate a number that depends on the level and on 
            #the shape of the ribbon
            factor = (self.x + self.y) % (l + 1)
            i = 0
            for _, v in enumerate(self.shape):
                i = 2 * i + v 
            factor = ((factor + i) * 79) % cmap.N
            #print("Factor = ", factor)
            #print("Max Len = ", MaxLength)             
            #c = cmap(factor/MaxLength)
            c = cmap(factor) #for integers cmap converts its range 
                                    #(given by cmap.N) to RGB colors.
                                    #for prism the range is 256
        else: 
            #calculate a number that depends on 
            #the shape of the ribbon
            i = 0
            for _, v in enumerate(self.shape):
                i = 2 * i + v 
            factor = (i * 79) % cmap.N
            #print("Factor = ", factor)
            #print("Max Len = ", MaxLength)             
            #c = cmap(i/MaxLength)
            c = cmap(factor) #for integers cmap converts its range 
                                    #(given by cmap.N) to RGB colors.
            
        
        rectangle = plt.Rectangle((self.x,self.y), 1, 1, fc=c ,ec="black")
        ax.add_patch(rectangle)
        
        sx = self.x
        sy = self.y
        for i in range(len(self.shape)):
            if self.shape[i] == 0:
                rectangle = plt.Rectangle((sx,sy), 2, 1, fc=c,ec="black")
                ax.add_patch(rectangle)
                if i > 0 and self.shape[i - 1] == 0:
                    line = plt.Line2D((sx, sx),(sy, sy + 1), color = c)
                    ax.add_line(line)
                elif i > 0 and  self.shape[i - 1] == 1:
                    line = plt.Line2D((sx, sx + 1),(sy, sy), color = c)
                    ax.add_line(line) 
                sx = sx + 1 
            else:
                rectangle = plt.Rectangle((sx,sy), 1, 2, fc=c,ec="black")
                ax.add_patch(rectangle) 
                if i > 0 and self.shape[i - 1] == 0:
                    line = plt.Line2D((sx, sx),(sy, sy + 1), color = c)
                    ax.add_line(line)
                elif i > 0 and  self.shape[i - 1] == 1:
                    line = plt.Line2D((sx, sx + 1),(sy, sy), color = c)
                    ax.add_line(line) 
                sy = sy + 1
            
            
        if block:            
            plt.show()
        else:
            plt.draw()
        return fig, ax
    
    
def squares2ribbon(squares):
    ''' build a ribbon from a collection of squares.
    
    parameters
    -------------
    squares: a list of tuples (x,y)
        represent a ribbon as a collection of squares. 
        they should be in right order. (either x or y increases at each 
        step.)
    '''
    x0 = squares[0][0]
    y0 = squares[0][1]
    x = x0
    y = y0
    shape = []
    for i in range(1, len(squares)):
        if squares[i][0] - x == 1:
            shape.append(0)
            x += 1
        else: 
            shape.append(1)
            y  += 1      
    ribbon = Ribbon(x0, y0, shape)
    return ribbon
        
'''
For testing methods
'''
def main():
    print('Ribbon methods are used')  
    '''
    ribbon = Ribbon(0, 0, [0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 1, 1, 1]) 
    _, ax1 = plt.subplots()
    ax1.set_xlim(0, 10)
    ax1.set_ylim(0, 10)
    print(ribbon)
    print(ribbon.squares)
    ribbon.draw(ax = ax1)
    '''
    
    '''
    checking squares to ribbon
    '''
    '''
    #squares = [(0,0), (0, 1), (1, 1)]
    squares = [(0, 2), (0, 3), (0, 4)]
    ribbon = squares2ribbon(squares)
    print(ribbon)
    _, ax2 = plt.subplots()
    ax2.set_xlim(0, 10)
    ax2.set_ylim(0, 10)
    ribbon.draw(ax = ax2)
    print(ribbon.contains(1, 4))
    '''
    
    ''' This is a couple of tests for the flip function 
    (seems to work OK)
    '''
    
    #a unit test 
    
    A = Ribbon(0, 0, [0, 0, 0, 1, 1, 0, 0, 1])
    B = Ribbon(2, 1, [1, 1, 0, 0, 1, 0, 0, 0])
    flag, A1, B1 = A.flip(B)
    print(flag)
    print(A1)
    print(B1)
    
    
    #some pictures
    from rgs.ribbons.Tiling import Tiling
    
    _, ax = plt.subplots(2, 2)
    
    #Case 1
    tiling = Tiling([A, B]) 
    print(tiling)
    tiling.draw(10, 10, ax = ax[0][0])
    
    if flag: 
        tiling = Tiling([A1, B1])
        tiling.draw(10, 10, ax = ax[0][1])
    
    #Case 2
    C = Ribbon(0, 0, [1, 1, 1, 1, 0])
    D = Ribbon(1, 2, [1, 0, 1, 1, 0])
    tiling = Tiling([C, D]) 
    print(tiling)
    tiling.draw(10, 10, ax = ax[1][0])
    
    flag, C1, D1 = C.flip(D)
    print(flag)
    print(C1)
    print(D1)
    if flag:
        tiling = Tiling([C1, D1])
        tiling.draw(10, 10, ax = ax[1][1])
        
    
    #simple cases
    
    #Case 3
    _, ax = plt.subplots(3, 2)
    A = Ribbon(0, 0, [0, 1])
    B = Ribbon(0, 1, [1, 0])
    
    tiling = Tiling([A, B]) 
    print(tiling)
    tiling.draw(10, 10, ax = ax[0][0])
    
    flag, A1, B1 = A.flip(B)
    print(flag)
    print(A1)
    print(B1)
    tiling = Tiling([A1, B1])
    tiling.draw(10, 10, ax = ax[0][1])
    
    
    #Case 4
    A = Ribbon(0, 0, [0, 0])
    B = Ribbon(0, 1, [0, 0])
    
    tiling = Tiling([A, B]) 
    print(tiling)
    tiling.draw(10, 10, ax = ax[1][0])
    
    flag, A1, B1 = B.flip(A)
    print(flag)
    print(A1)
    print(B1)
    tiling = Tiling([A1, B1])
    tiling.draw(10, 10, ax = ax[1][1])
    
    
    #Case 5
    A = Ribbon(0, 0, [1, 1])
    B = Ribbon(1, 0, [0, 1])
    
    tiling = Tiling([A, B]) 
    print(tiling)
    tiling.draw(10, 10, ax = ax[2][0])
    
    flag, A1, B1 = B.flip(A)
    print(flag)
    print(A1)
    print(B1)
    tiling = Tiling([A1, B1])
    tiling.draw(10, 10, ax = ax[2][1])
    
    
    
    '''
    checking the coloring scheme
    '''
    '''
    ribbon = Ribbon(0, 0, [0, 1]) 
    _, ax = plt.subplots()
    ax.set_xlim(0, 10)
    ax.set_ylim(0, 10)
    print(ribbon)
    print(ribbon.squares)
    ribbon.draw(ax = ax, colorByLength = False)
    
    ribbon = Ribbon(0, 0, [1, 1]) 
    _, ax = plt.subplots()
    ax.set_xlim(0, 10)
    ax.set_ylim(0, 10)
    print(ribbon)
    print(ribbon.squares)
    ribbon.draw(ax = ax, colorByLength = False)
    
    ribbon = Ribbon(1, 0, [1, 1]) 
    _, ax = plt.subplots()
    ax.set_xlim(0, 10)
    ax.set_ylim(0, 10)
    print(ribbon)
    print(ribbon.squares)
    ribbon.draw(ax = ax, colorByLength = False)
    
    ribbon = Ribbon(2, 0, [1, 1]) 
    _, ax = plt.subplots()
    ax.set_xlim(0, 10)
    ax.set_ylim(0, 10)
    print(ribbon)
    print(ribbon.squares)
    ribbon.draw(ax = ax, colorByLength = False)
    '''
    plt.show()
    
    
if __name__ == "__main__":
    main() 
    
    


'''
A package to manipulate and display some random structures,
including meander systems, planar triangulations, and ribbon tilings


Created on May 8, 2021

@author: vladislavkargin
'''

'''
#I prefer blank __init__.py

from . import mndrpy
from . import pmaps
from . import ribbons
'''


'''
Created on Jan 20, 2021

@author: vladislavkargin
'''

import numpy as np
import matplotlib.pyplot as plt
from itu.algs4.graphs import digraph as dgr


import rgs.ribbons.Ribbon as rb
import rgs.ribbons.Utility as ut

class Tiling(object):
    '''
    Realizes a tiling, which is a list of non-intersecting ribbons.

    '''


    def __init__(self, ribbons):
        '''
        Constructor
        Parameters
        ---------------
        ribbons: a list of ribbons
        
        '''
        self.ribbons = ribbons
    
    def findRibbon(self, x, y):
        ''' find the ribbon that covers square (x, y) '''
        for ribbon in self.ribbons:
            if ribbon.contains(x,y):
                return ribbon
        return None
    
        
    def flip(self, i, j): 
        ''' flip tiles number i and number j and returns true, if possible,
        do nothing and returns false if not possible.
        
        Parameters
        --------------
        i, j: integers 
            the indices of ribbons to flip
        
        Returns:
            True if successful
            False if impossible
        '''
        
        A = self.ribbons[i]
        B = self.ribbons[j] 
        #print(A)
        #print(B)
        flag, A1, B1 = A.flip(B)
        #print(A1)
        #print(B1)
        if flag:
            self.ribbons[i] = A1
            self.ribbons[j] = B1
        return flag
          
        
    def __str__(self):
        ''' Returns a string representation of this tiling
        return the graph'''
        s = ''
        for ribbon in self.ribbons:
            s = s + '\n' + str(ribbon)
        return s 
    
    def draw(self, M, N, ax = None, 
                colormap = 'prism', colorType = "Shape",
                offset = [0, 0],
                MaxLength = None,
                block = False):
        ''' draw the tiling 
        Parameters
        --------------
        M, N : integers 
            the height and the width of the plot 
        ax : matplotlib axis
            axis in which to draw (default = None)
        colormap : 'prism' 
            the name of the coloring scheme ("palette"), used to color the tiling,
            default = "prism"
            other possibilities: "jet"and many others, see matplotlib docs.
        colorType : string
            defines the type of coloring. Possible choices:
            "Shape" -- color only depends on shape
            "ShapePosition" -- color depends on shape and level (mod n)
            "Length" -- color only depends on length of the ribbon
            (default = "Shape")
        offset: a list of two integers
            sets the origin of the axis at 0 - offset[0], 0 - offset[1]
            (default = [0, 0])
        MaxLength : integer 
            a fine-tuning parameter used in the choice of coloring (default = None)
        block: boolean
            if True, stops the execution and draws the graph (default = False)
        
        
        
        '''
        if (ax == None):
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
            
        ax.set_xlim(0 - offset[0], N)
        ax.set_xticks(list(range(-offset[0],N)))
        ax.set_ylim(0 - offset[1], M)
        ax.set_yticks(list(range(-offset[1], M)))
        
        if MaxLength == None:
            if colorType == "Length":
                MN = 3 * int(np.log2(max([M, N])))
            else:
                MN = None #Let the ribbon itself decide what color to choose
                        #based on position and on the length
        else:
            MN = MaxLength
        #print(MN)
        for ribbon in self.ribbons:
            ribbon.draw(ax, MaxLength = MN, colorType = colorType,
                        colormap = colormap)
        
        #ax.grid()
        if block:            
            plt.show()
        else:
            plt.draw()
        return fig, ax


def fromString(s):
    '''
    initializes a tiling from a string like the following
    '(0,0), [1, 1]
    (2,1), [1, 0]
    (1,0), [1, 1]'
     '''
    #TODO
    pass

def standardAztec2(M):
    '''
    generate a standard tiling of an M-by-M Aztec Diamond by dominos
    '''
    ribbons = []
    
    for i in range(M): #the first half of the diamond
        for j in range(M - i): 
            ribbon = rb.Ribbon(i + j, i - j, [1])
            ribbons.append(ribbon)
    
    for i in range(M): #the second half of the diamond
        for j in range(i + 1): 
            ribbon = rb.Ribbon(M + j , - M + 2 * i - j + 1, [1])
            ribbons.append(ribbon)
            #print(str(ribbon))
    t = Tiling(ribbons)
    return t
    

def standardT(n, M, N = None):
    '''
    generate a standard tiling of Mn - by - N rectangle (Mn - height,
    N - width)
    by vertical n-ribbons. If N = None, returns a square 
    with side M * n
    '''
    if N == None:
        N = n * M 
        
    ribbons = []
    #This should be changed. we need to add ribbons in standard order. 
    levels, vertex2level = _getHelpersRect(n, M, N)
    #print("vertex2level = ", vertex2level)
    for v, level in enumerate(vertex2level):
        position = levels[level].index(v)
        if level < M * n:
            #print("v = ", v, " level = ", level)
            #print("position = ", position)
            x = position * n + level % n
            y = level - x 
            #print("x = ", x, " y = ", y)
            ribbon = rb.Ribbon(x, y, [1] * (n - 1))
            ribbons.append(ribbon)
        else:
            x = position * n + level - (M - 1) * n
            y = level - x 
            ribbon = rb.Ribbon(x, y, [1] * (n - 1))
            ribbons.append(ribbon)
    t = Tiling(ribbons)
    return t

def randomT(n, M, N, ITER = 10, seed = None):
    '''
    generate a random tiling of Mn - by - N rectangle (Mn - height,
    N - width)
    by using a version of Markov Chain algorithm with flips
    
    Parameters
    ------------
    n: integer
        length of tile
    M, N: integers
        The region is Mn - by - N rectangle
    ITER: integer
        parameter that determines how long to run the algorithm
        typically every tile is touched around ITER times
    seed: integer
        the seed of the random generator, for reproducibility,
        default = None
    
    Returns
    ------------
    random ribbon tiling
    
    '''
    #There are some errors 
    
    if (seed != None):
        np.random.seed(seed)
    else:
        np.random.seed()
        
    t = standardT(n, M, N)
    V = len(t.ribbons)
    #for v in range(V):
    #    print(t.ribbons[v])
    helpers = _getHelpersRect(n, M, N)
    #print("Herlpers =", helpers)
    for count in range(ITER):
        for v in range(V):
            #print("v = ", v)
            flips = ut.findFlips(n, v, t, helpers)
            if len(flips) > 0:
                i = np.random.randint(len(flips))
                w = flips[i]
                #print("w = ", w)                
                t.flip(v,w)
    return t

def _getHelpersRect(n, M, N):
    '''
    create helper structures for a rectangle with number of rows M * N 
    and number of columns N
    
    (In this version I do not create a digraph.)
    
    The vertices are arranged in layers according to the level and ordered 
    inside the layer. 
    
    Parameters
    -------------
    n: integer
       the size of the ribbon
    M: integer
       determines the number of rows(the actual size is M * n)
    N: integer 
        the number of columns
       
    Returns
    --------------
    (levels, level_lengths, integer2level) where
       
    levels: a list of integer lists
          every inside list represents vertices in the level
    
    integer2level: a list of integers
            the k-th element is the level of vertex k.
       
    
    '''
    
    #digraph = dgr.Digraph(M * N * n)
    levels = []
    #create a list of levels
    nlevels = (M - 1) * n + N
    level_lengths = [0] * nlevels
    
    #make an auxiliary tiling
    ribbons = []
    for i in range(N):
        for j in range(M):
            ribbon = rb.Ribbon(i, j*n, [1] * (n - 1))
            ribbons.append(ribbon)
    st = Tiling(ribbons)
    #print(len(ribbons))
    #print("Auxiliary = ", st)
    
    for ribbon in st.ribbons:
        l = ribbon.level()
        level_lengths[l] += 1
    #print(level_lengths)
    x = 0
    for l in range(nlevels):
        level = list(range(x, x + level_lengths[l]))
        levels.append(level)
        x = x + level_lengths[l]
    #print(levels)

    vertex2level = [-1] * M * N                    
    for l in range(nlevels):
        for i in range(level_lengths[l]):
            vertex2level[levels[l][i]] = l 
         
    return levels, vertex2level


def _getHelpers(n, M):
    '''
    create a generic digraph of a tiling for a square
    with side M * n and other helper structures
    
    The generic means that only the forced edges are included.
    (and not all of them, only those that generate the structure of the graph)
    in particular, those between vertices in the same level are not included.
    
    The vertices are arranged in layers according to the level and ordered 
    inside the layer. 
    
    Parameters
    -------------
    n: integer
       the size of the ribbon
    M: integer
       the size of the square side (actual size is M * n)
       
    Returns
    --------------
    (digraph, levels, level_lengths, integer2level) where
    
    digraph : itu.algs4.Digraph 
        digraph of the tiling where only forced edges are included
        NOTE: I use the convention from Yinsong's algorithm, v -> w 
        if w is on the right of v
    
    levels: a list of integer lists
          every inside list represents vertices in the level
          
    level_lengths: list of integers
        the length of each level
    
    integer2level: a list of integers
            the k-th element is the level of vertex k.
       
    
    '''
    N = M 
    
    digraph = dgr.Digraph(M * N * n)
    levels = []
    #create a list of levels
    st = standardT(n, M)
    nlevels = (N + M - 1) * n
    level_lengths = [0] * nlevels
    for ribbon in st.ribbons:
        l = ribbon.level()
        level_lengths[l] += 1
    #print(level_lengths)
    x = 0
    for l in range(nlevels):
        level = list(range(x, x + level_lengths[l]))
        levels.append(level)
        x = x + level_lengths[l]
    #print(levels)
    #Now we want make a digraph
    for l in range(nlevels):
        level = levels[l]
        for i, v in enumerate(level): 
            if l < (M - 1) * n: #lower levels
                if l < n:
                    w = levels[l + n][i + 1]
                    digraph.add_edge(v, w)
                else:
                    if i + 1 < level_lengths[l + n]:
                        w = levels[l + n][i + 1]
                        digraph.add_edge(v, w)
                    if i < level_lengths[l - n]:
                        #print("v = ", v, "i = ", i, "l = ", l)
                        w = levels[l - n][i]
                        digraph.add_edge(v, w)
            elif l < M * n: #middle level
                    if i < level_lengths[l + n]:
                        w = levels[l + n][i]
                        #print("v = ", v, "i = ", i, "l = ", l, "w = ", w)
                        digraph.add_edge(v, w)
                    if i < level_lengths[l - n]:
                        #print("v = ", v, "i = ", i, "l = ", l)
                        w = levels[l - n][i]
                        digraph.add_edge(v, w)
            else: #upper levels
                if l >= nlevels - n:
                    w = levels[l - n][i + 1]
                    digraph.add_edge(v, w)
                else:
                    if i + 1 < level_lengths[l - n]:
                        w = levels[l - n][i + 1]
                        digraph.add_edge(v, w)
                    if i < level_lengths[l + n]:
                        #print("v = ", v, "i = ", i, "l = ", l)
                        w = levels[l + n][i]
                        digraph.add_edge(v, w)   
    vertex2level = [-1] * M * N * n                    
    for l in range(nlevels):
        for i in range(level_lengths[l]):
            vertex2level[levels[l][i]] = l 
         
    return digraph, levels, level_lengths, vertex2level

def buildFromTilingSeq(seq, n, M):
    '''build a tiling of a rectangle from a tiling sequence
    
    Parameters
    -----------
    seq: a list of integers, 
    represents the permutation of tiles/graph vertices [0, len(seq) - 1]
    which induces an acyclic orientation on graph in agreement 
    with the forced partial orientation. 
    
    n: integer
       the size of the ribbon
    M: integer
       the size of the square side (actual size is M * n)
    
    Returns: a tiling 
    '''
    (_, levels, _, vertex2level) = _getHelpers(n,M)
    
    ribbons = []
    
    for i, v in enumerate(seq): #for every vertex build a tile
        #print("v = ", v)
        squares = []
        l = vertex2level[v]
        for lt in range(l, l+n): #for every level of the tile
            #print("lt = ", lt)
            #count the number of vertices which are comparable
            #and precede v in the sequence
            counter = 0
            for j in range(lt - n + 1, lt + 1): #comparable levels
                if j >= 0 and j < len(levels):
                    #print("j = ", j)
                    for w in levels[j]:
                        if w in seq and seq.index(w) < i:
                            #print("w = ", w)
                            counter += 1
            #print("counter = ", counter)               
            if (lt < M * n):                
                square = (counter, lt - counter)               
            else:
                square = (counter + lt - M * n + 1, M * n - counter - 1)
            #print("square = ", square)
            squares.append(square)
        #print(squares)
        ribbon = rb.squares2ribbon(squares)
        ribbons.append(ribbon)
    tiling = Tiling(ribbons)
    return tiling


def makeRandomTS(n, M):
    '''create a random tiling sequence for a square with the
    side n * M
    
    The tiling sequence is build using Yinsong Chen's algorithm. 
    It is NOT uniform on the space of all tiling sequences. 
    
    Parameters
    -----------
    n: integer
       the size of the ribbon
    M: integer
       the size of the square side (actual size is M * n)
       
    Returns 
    -------
    
    tiling sequence: integer array
        this sequence can be used to create 
        a tiling using buildFromTilingSequence
    
    '''
    
    (digraph, _, _, vertex2level) = _getHelpers(n,M)
    #initiate a sequence
    rev_digraph = digraph.reverse()
    ts = list(np.random.permutation(n))
    #print(ts)
    
    num_vertices = M * M * n
    for v in range(n, num_vertices):
        #print("v = ", v)
        #find bounds from the level below
        lb = -1
        ub = num_vertices
        for w in digraph.adj(v):
            if vertex2level[w] == vertex2level[v] - n:
                ub = w
        for w in rev_digraph.adj(v):
            if vertex2level[w] == vertex2level[v] - n:
                lb = w 
        #print(f'(lb, ub) = ({lb}, {ub})')
        #find the positions of all the points 
        #in the sequence which are between the bounds and
        #are comparable with v (the list can be empty)
        l_index = -1 #location of lower bound
        u_index = len(ts) #location of upper bound vertex
        positions = []
        if lb in ts:
            l_index = ts.index(lb)
        if ub in ts:
            u_index = ts.index(ub)
        #print(f'(l_index, u_index) = ({l_index}, {u_index})')
        for i in range(l_index + 1, u_index):
            x = ts[i]
            if vertex2level[x] > vertex2level[v] - n:
                positions.append(i)
        #print("positions = ", positions)
        if len(positions) == 0:
            ts.insert(l_index + 1, v)
        else:
            j = np.random.randint(len(positions) + 1)
            #print("j = ", j)
            if j < len(positions):
                ts.insert(positions[j], v)
            else:
                ts.insert(positions[-1] + 1, v)  #insert after the 
                                                #last one in positions
        #print("ts = ", ts)                                         
    return ts

def _getHelpersAztec2(M):
    '''
    create a generic digraph of a domino tiling for an Aztec
    Diamond with side M and other helper structures.
    
    
    The generic means that only the forced edges are included.
    (and not all of them, only those that generate the structure of the graph)
    in particular, those between vertices in the same level are not included,
    except perhaps for the first level
    
    The vertices are arranged in layers according to the level and ordered 
    inside the layer. 
    
    Parameters
    -------------
    M: integer
       the size of the square side (actual size is M * n)
       
    Returns
    --------------
    (digraph, levels, level_lengths, integer2level) where
    
    digraph : itu.algs4.Digraph 
        digraph of the tiling where only forced edges are included
        NOTE: I use the convention from Yinsong's algorithm, v -> w 
        if w is on the right of v
    
    levels: a list of integer lists
          every inside list represents vertices in the level
          
    level_lengths: list of integers
        the length of each level
    
    integer2level: a list of integers
            the k-th element is the level of vertex k.
       
    
    '''
    digraph = dgr.Digraph(M * (M + 1))
    levels = []
    #create a list of levels
    az = standardAztec2(M)
    nlevels = 2 * M
    level_lengths = [0] * nlevels 
    for ribbon in az.ribbons:
        l = ribbon.level()
        #print("l = ", l)
        level_lengths[l] += 1
    #print(level_lengths)
    x = 0
    for l in range(nlevels):
        level = list(range(x, x + level_lengths[l]))
        levels.append(level)
        x = x + level_lengths[l]
    #print(levels)
    #building digraph
    #first, we go along all vertices in even levels (except the last one)
    for l in range(0, 2 * (M - 1), 2):
        for i in range(level_lengths[l]):
            v = levels[l][i]
            if i > 0 and i < level_lengths[l] - 1:
                #print("l = ", l, "i = ", i)
                #print(levels[l + 2])
                w1 = levels[l + 2][i - 1]
                w2 = levels[l + 2][i]
                #digraph.add_edge(v,w1)
                #digraph.add_edge(w2,v)
                digraph.add_edge(v,w2)
                digraph.add_edge(w1,v)
            elif i == 0:
                w2 = levels[l + 2][i]
                #digraph.add_edge(w2,v)
                digraph.add_edge(v, w2)
            else:
                w1 = levels[l + 2][i - 1]
                #digraph.add_edge(v,w1)
                digraph.add_edge(w1,v)
        #additional adjustment for level 0
        #if l == 0:
        #    for i in range(1, level_lengths[l]):
        #        digraph.add_edge(i - 1, i)
                
    #second, we will go along the odd levels (except the last one)
    for l in range(1, 2 * (M - 1), 2):
        for i in range(level_lengths[l]):
            v = levels[l][i]
            #print("l = ", l, "i = ", i)
            #print(levels[l + 2])
            w1 = levels[l + 2][i]
            w2 = levels[l + 2][i + 1]
            #digraph.add_edge(v,w1)
            #digraph.add_edge(w2,v)
            digraph.add_edge(v, w2)
            digraph.add_edge(w1,v)
    #print(digraph)
    vertex2level = [-1] * M * (M + 1)                  
    for l in range(nlevels):
        for i in range(level_lengths[l]):
            vertex2level[levels[l][i]] = l 
    #print(vertex2level)
    return digraph, levels, level_lengths, vertex2level

def buildAztec2FromTilingSeq(seq, M):
    '''build a domino tiling of an Aztec diamond  from a tiling sequence
    
    Parameters
    -----------
    seq: a list of integers, 
    represents the permutation of tiles/graph vertices [0, len(seq) - 1]
    which induces an acyclic orientation on graph in agreement 
    with the forced partial orientation. 

    M: integer
       the size of the Aztec diamond 
    
    Returns: a tiling 
    '''
    #TODO (smth not OK)
    (_, levels, _, vertex2level) = _getHelpersAztec2(M)
    
    ribbons = []
    n = 2 #we are working with dominos
    for i, v in enumerate(seq): #for every vertex build a tile
        #print("v = ", v)
        squares = []
        l = vertex2level[v]
        for lt in range(l, l+n): #for every level of the tile
            #print("lt = ", lt)
            #count the number of vertices which are comparable
            #and precede v in the sequence
            counter = 0
            for j in range(lt - n + 1, lt + 1): #comparable levels
                if j >= 0 and j < len(levels):
                    #print("j = ", j)
                    for w in levels[j]:
                        if w in seq and seq.index(w) < i:
                            #print("w = ", w)
                            counter += 1
            #print("counter = ", counter)
            if lt < 2 * M: 
                square = (lt//2 + counter, lt - (lt//2 + counter))   
            else:
                square = (l//2 + counter + 1, lt - (l//2 + counter + 1))   
            # we should work with odd and even l separately
            #if l % 2 == 0:
            #    square = (l//2 + counter, lt - (l//2 + counter))
            #else: 
            #    square = (l//2 + counter, lt - (l//2 + counter))
            #print("square = ", square)
            squares.append(square)
        #print(squares)
        ribbon = rb.squares2ribbon(squares)
        ribbons.append(ribbon)
    tiling = Tiling(ribbons)
    return tiling

def makeRandomTS_Aztec2(M):
    '''create a random domino tiling sequence for an Aztec Diamond with the
    side M
    
    Parameters
    -----------
    M: integer
       the size of the Aztec Diamond side 
       
    Returns 
    -------
    
    tiling sequence: integer array
        this sequence can be used to create 
        a tiling using buildAztecFromTilingSeq
    
    '''
    #the specific 
    (digraph, _, _, vertex2level) = _getHelpersAztec2(M)
    #initiate a sequence
    rev_digraph = digraph.reverse()
    ts = list(range(M))
    #print(ts)
    
    num_vertices = digraph.V()
    n = 2
    for v in range(M, num_vertices):
        #print("v = ", v)
        #find bounds from the level below
        lb = -1 #default values for upper and lower bounds
        ub = num_vertices
        for w in digraph.adj(v):
            if vertex2level[w] == vertex2level[v] - n:
                ub = w
        for w in rev_digraph.adj(v):
            if vertex2level[w] == vertex2level[v] - n:
                lb = w 
        #print(f'(lb, ub) = ({lb}, {ub})')
        #find the positions of all the points 
        #in the sequence which are between the bounds and
        #are comparable with v (the list can be empty)
        l_index = -1 #location of lower bound
        u_index = len(ts) #location of upper bound vertex
        positions = []
        if lb in ts:
            l_index = ts.index(lb)
        if ub in ts:
            u_index = ts.index(ub)
        #print(f'(l_index, u_index) = ({l_index}, {u_index})')
        for i in range(l_index + 1, u_index):
            x = ts[i]
            if vertex2level[x] > vertex2level[v] - n:
                positions.append(i)
        #print("positions = ", positions)
        if len(positions) == 0:
            ts.insert(l_index + 1, v)
        else:
            j = np.random.randint(len(positions) + 1)
            #print("j = ", j)
            if j < len(positions):
                ts.insert(positions[j], v)
            else:
                ts.insert(positions[-1] + 1, v)  #insert after the 
                                                #last one in positions
        #print("ts = ", ts)                                         
    return ts



'''
For testing methods
'''
def main():
    print('Ribbon tiling methods are used')  
    
    ''' testing basic construction '''
    '''
    ribbon1 = rb.Ribbon(0, 0, [0, 1])
    ribbon2 = rb.Ribbon(0, 1, [1, 0])
    
    tiling = Tiling([ribbon1, ribbon2]) 
    _, ax1 = plt.subplots()
    print(tiling)
    tiling.draw(10, 10, ax = ax1, MaxLength = 3, 
                colorByLength = False, colormap = "prism")
    '''
    
    
    A = rb.Ribbon(0, 0, [0, 0, 0, 1, 1, 0, 0, 1])
    B = rb.Ribbon(2, 1, [1, 1, 0, 0, 1, 0, 0, 0])
    
    tiling = Tiling([A, B]) 
    print(tiling)

    '''
    (0,0), [1, 1]
    (2,1), [1, 0]
    (1,0), [1, 1]
    (1,3), [1, 1]
    (2,0), [0, 1]
    (2,4), [1, 0]
    (0,3), [1, 1]
    (2,3), [0, 1]
     '''   
    
    '''testing building a standard tiling'''
    
    n = 3
    M = 2
    N = 3
    #t = standardT(n, M, N)
    #print(t)
    _, ax2 = plt.subplots(1,1)
    #t.draw(M * n, N, ax = ax2)
    
    #let us look at some helper structures
    '''
    levels, vertex2level = _getHelpersRect(n, M, N)
    print(levels)
    print(vertex2level)
    '''
    
    '''testing building a random tiling'''
    n = 5
    M = 1
    N = 30
    helpers = _getHelpersRect(n, M, N)
    t = randomT(n, M, N, ITER = 100)
    #print(t)
    t.draw(M * n, N, ax = ax2)
    flips = ut.findFlips(n, 4, t, helpers)
    
    A = t.ribbons[4]
    print("A = ", A)
    print(flips)
    #B = t.ribbons[5]
    #(flag, C, D)= A.flip(B)
    #print(flag)
    #_, ax = A.draw()
    #B.draw(ax = ax)
    #print(flips)
    
    
    
    #t.flip(4,5)
    #t.draw(M * n, N, ax = ax2[1])
    
    '''
    Testing Yinsong's algorithm for the random generation of a rectangle tiling
    '''
    '''
    n = 3
    M = 3
    digraph, levels, level_lengths, vertex2level = getHelpers(n, M)
    
    print(digraph)
    print(level_lengths)
    print(levels)
    print(vertex2level)
    
    #seq = [2, 1, 0]
    #seq = [1, 2, 0]
    #seq = [7, 2, 5, 3, 0, 1, 4, 6]
    #num_vertices = M * M * n
    #seq = list(range(3, -1, -1))
    #print(seq)
    seq = [5, 7, 3, 1, 0, 2, 9, 10, 11, 6, 4, 8]
    tiling = buildFromTilingSeq(seq, n, M)
    tiling.draw(M * n, M * n)
    
    
    ts = makeRandomTS(n, M)
    tiling = buildFromTilingSeq(ts, n, M)
    tiling.draw(M * n, M * n)
    '''
    
    
    '''
    Testing Yinsong's algorithm for the random generation of
     an Aztec Diamond tiling. Here it is clear that the tiling is not 
     sampled from the uniform distribution. 
    '''
    
    '''
    M = 4
    t = standardAztec2(M)
    t.draw(M + 2, 2 * M, offset = (0, M))
    '''
    
    '''
    #getHelpersAztec2(M)
    M = 40
    #seq = [0, 4, 1, 9, 8, 6, 5, 3, 2, 10, 7, 11]
    #seq = [0, 4, 9, 6, 3, 1, 10, 7, 11, 8, 5, 2]
    #seq =  [19, 14, 18, 3, 1, 9, 13, 7, 5, 0, 4, 2, 8, 11, 15, 17, 6, 10, 12, 16]
    seq = makeRandomTS_Aztec2(M)
    print("seq = ", seq)
    tiling = buildAztec2FromTilingSeq(seq, M)
    #print("tiling = ", tiling)
    tiling.draw(M + 2, 2 * M + 1, offset = (1, M))
    
    #1) is it possible that 3 is before 0 in tiling with M = 3?
    '''

    
    
    
    plt.show()
    
    
    
if __name__ == "__main__":
    main() 


'''
Created on Jul 23, 2020

@author: vladislavkargin
'''

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib import cm

import rgs.mndrpy.Pairing as pr
#import rgs.mndrpy.Utility as ut
import rgs.mndrpy.DyckPaths as dp

class Meander(object):
    '''
    classdocs
    '''


    def __init__(self, uPairing, dPairing, uprngArray = None,
                    dprngArray = None):
        '''
        Constructor
        Makes the meander from upper and lower pairings
        arguments:
        uPairing, dPairing -- upper and lower parings (as objects or lists)
        
        The other pair of arguments is obsolete (poor design),
        kept only for compatibility with some other functions. 
        uprgnArray, dprngArray -- upper and lower pairings (as lists)
        '''
        if type(uPairing) is list:
            self.uPairing = pr.Pairing(prngArray = uPairing)
        elif uPairing == None:
            self.uPairing = pr.Pairing(prngArray = uprngArray)
        else: 
            self.uPairing = uPairing
            
        if type(dPairing) is list:
            self.dPairing = pr.Pairing(prngArray = dPairing)
        elif uPairing == None:
            self.dPairing = pr.Pairing(prngArray = dprngArray)
        else: 
            self.dPairing = dPairing
        
        
    def __str__(self):
        return str(self.uPairing) + ", " + str(self.dPairing)
    
    def __eq__(self, other):
        '''Overrides the default implementation'''
        if isinstance(other, Meander):
            if self.uPairing == other.uPairing and self.dPairing == other.dPairing: 
                return True
            else: 
                return False
        return False
    
    def __hash__(self):
        return hash(self.uPairing) + 17 * hash(self.dPairing)
    
    def size(self):
        ''' returns n for the meander system on 2n bridges '''
        return self.uPairing.size()
     
    def reflected(self):
        '''returns a meander with lower and upper pairings exchanged
        '''     
        new_mndr = Meander(uPairing = self.dPairing, dPairing = self.uPairing)
        return new_mndr
    
    def draw(self, ax = None, drawCycles = False, palette = "jet", 
             asPolygon = False):
        if ax == None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
        fig = ax.figure
        self.uPairing.draw(ax = ax, asPath = asPolygon)
        self.dPairing.draw(ax = ax, up = False, asPath = asPolygon)
        if asPolygon:
            ax.grid(which='both')
        if drawCycles and not asPolygon:
            colormap = cm.get_cmap(palette, 128)
            cycles, _ = self.findCycles()
            for i in range(len(cycles)):
                col = colormap(i/len(cycles))
                #print(i, col)
                self.drawCycle(cycles[i], ax, color = col) 
        return fig, ax
    
    def drawCycle(self, cycle, ax = None, color = "red"):
        if ax == None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            width = len(self.uPairing.prng)
            ax.set_xlim(0, width)
            ax.set_ylim(-width/2, width/2)
        for c, i in enumerate(cycle):
            j = cycle[(c + 1) % len(cycle)]
            xy = (i + (j - i)/2, 0)
            w = np.abs((j - i)/2)
            if c % 2 == 0:
                theta1 = 0.0
                theta2 = 180.0
            else:
                theta1 = 180.0
                theta2 = 360.0
            arc = mpatches.Arc(xy, 2*w, 2*w, 0, theta1, theta2, color = color)
            ax.add_patch(arc)
    
    def isProper(self):
        '''checks if this system has only one cycle'''  
        N = len(self.uPairing.prng)
        #print(N)
        next_v = self.uPairing.prng[0]
        up = False 
        counter = 0    
        while next_v != 0 and counter <= 2*N:
            counter = counter + 1
            if up == True:
                next_v = self.uPairing.prng[next_v]
                up = False
            else:
                next_v = self.dPairing.prng[next_v]
                up = True
        if counter == N - 1:
            return True
        else:
            return False
        
    def  isIrreducible(self):
        '''checks if the meander is irreducible. (NB. The definition is different
        from irreducible meanders in the sense of Zvonkin-Lando, so it is
        more of exploratory nature.)'''
        if self.isProper():
            return True   
        cycles, _ = self.findCycles()
        for cycle in cycles:
            if len(cycle) == max(cycle) - min(cycle) + 1:
                return False
        return True
    
        
    def findCycles(self):
        ''' finds all cycles in the meander and stores them in two structures
        cycles - which is an array of arrays that represent cycles
        cyclesInd - which is an 2n-array with each element marked by 
        the number of the cycle that it belongs to.
        '''
        N = len(self.uPairing.prng)
        c = 0
        up = True #shows which pairing we are going to use in tracking 
        cycles = []
        cyclesInd = [-1] * N
        
        for i in range(N):
            if cyclesInd[i] == -1:
                c = c + 1
                cyclesInd[i] = c
                cycle = [i]
                if up:
                    next_v =  self.uPairing.prng[i]
                    up = False
                else:
                    next_v =  self.dPairing.prng[i]
                    up = True
                #the following cycle can go in infinite loop if one of the
                #pairing is invalid for this reason we break it 
                # if it go for more than N iterations.
                # The result is invalid and likely to break the following programs
                counter = 0
                while (next_v != i and counter <= 2 * N):
                    cyclesInd[next_v] = c
                    cycle.append(next_v)
                    if up:
                        next_v =  self.uPairing.prng[next_v]
                        up = False
                    else:
                        next_v =  self.dPairing.prng[next_v]
                        up = True
                    counter = counter + 1
                cycles.append(cycle)
                if counter > 2 * N:
                    print( "findCycles Warning:")
                    print(" An infinite loop in find components was broken")
                    print(" A likely problem with validity of one of the pairings")
        return cycles, cyclesInd
     
    def calcCycleLengths(self):
        ''' caclulate the lengths of all cycles in the meander'''
        (cycles, _) = self.findCycles()
        lengths = []
        for cycle in cycles:
            lengths.append(len(cycle))
        return lengths
        
def areaMeander(mndr):
    '''calculates the area of a meander system defined as the area 
    of the polygone that has the Dyck path for the upper pairing
    as its upper boundary and the reflected Dyck path for the lower
    pairing as its lower boundary. The size of the meander is subtracted
    for normalization, so the area of the standard meander system is 0'''
    area = dp.area(pr.prng2path(mndr.uPairing))
    area += dp.area(pr.prng2path(mndr.dPairing))
    #area += mndr.size()
    return area
         
def randomMeander(n, w = None, seed = None):
    ''' generates a uniformly random meander on half-length n.
    If 2-tuple w != None, generates a meander with pairings distributed 
    according to weights w1 and w2'''
    if seed != None:
        np.random.seed(seed)
    if w == None:
        uP = pr.randomPairing(n)
        dP = pr.randomPairing(n)
    else:
        uP = pr.randomPairing(n, w[0])
        dP = pr.randomPairing(n, w[1])
    mndr = Meander(uP, dP)
    return mndr

def randomCrossSemiMeander(n, seed = None):
    '''generate a random crossing semi-meander'''
    if seed != None:
        np.random.seed(seed)
    uP = pr.randomCrossPairing(n)
    dP = pr.rainbowPairing(n)
    mndr = Meander(uP, dP)
    return mndr
    

def randomComb(n, seed = None):
    ''' generates a random comb meander of half-length n'''
    if seed != None:
        np.random.seed(seed)
    uP = pr.randomPairing(n)
    dP = pr.standardPairing(n)
    mndr = Meander(uP, dP)
    return mndr

def randomSemiMeander(n, seed = None):
    ''' generates a random meander of half-length n that have a random pairing
    on top and the rainbow pairing in the bottom'''
    
    if seed != None:
        np.random.seed(seed)
    uP = pr.randomPairing(n)
    dP = pr.rainbowPairing(n)
    mndr = Meander(uP, dP)
    return mndr

def rainbowMeander(uWidths, dWidths):
    '''
    generates a rainbow meander.
    Parameters:
    uWidths, dWidths -- sizes of the meanders on the top and on the bottom.
    returns a meander
    '''  
    N = sum(uWidths)
    M = sum(dWidths)
    if M != N:
        print("Rainbow meander says: the total sizes of upper and lower meanders")
        print("should be the same. N = ", N, " and M = ", M)
        return 
    uP = pr.rainbows(uWidths)
    dP = pr.rainbows(dWidths)
    return Meander(uP, dP)

def smallRainbowsMeander(amounts, sizes):
    '''
    creates a random meander whose pairings are collections of small rainbows
    '''
    uP = pr.smallRainbows(amounts, sizes)
    dP = pr.smallRainbows(amounts, sizes)
    return Meander(uP, dP)

def drawAsPolygon(mndr):
    ''' draws meander as polygon. The upper pairing is represented
    by a Dyck path above the diagonal y = x, and the lower pairing
    is represented by a Dyck path below the main diagonal. 
    '''
    upath = pr.prng2path(mndr.uPairing)
    ax = dp.plotDyckPath(upath)
    dpath = pr.prng2path(mndr.dPairing)
    ax = dp.plotDyckPath(dpath, ax = ax, method = "lowerLatticePath")
    return ax    
    
'''
For testing methods
'''
def main():
    
    #n = 100
    #seed = None    
    #mndr = randomMeander(n, seed = seed)
    #mndr = randomComb(n, seed = seed)
    #mndr = randomSemiMeander(n, seed = seed)
    #mndr = randomMeander(n, w = ([1,1,1],[1,1,1]), seed = seed)
    #mndr = randomCrossSemiMeander(n)
    amounts = [20, 20]
    sizes = [1, 5]
    mndr = smallRainbowsMeander(amounts, sizes)
    
    print(str(mndr))
    _, ax = mndr.draw(drawCycles = True)
    
    (cycles, cyclesInd) = mndr.findCycles()
    print("Cycles = " + str(cycles))
    print("CyclesInd = " + str(cyclesInd))
    
    lengths = mndr.calcCycleLengths()
    print("Cycle lengths = " + str(lengths))
    
    maxL = max(lengths)
    i = lengths.index(maxL)
    
    print("The length of the largest cycle is " + str(maxL))
    
    mndr.drawCycle(cycles[i], ax = ax)
    
    print("Hash = ", hash(mndr))
    print("Meander? ", mndr.isProper())
    print("Irreducible? ", mndr.isIrreducible())
    
    
    
    ''' Picture for the paper about meanders'''
    
    uWidths = [15, 5, 11]
    dWidths = [18, 13]
    mndr = rainbowMeander(uWidths, dWidths)
    _, ax = mndr.draw(drawCycles = True)  
    (cycles, cyclesInd) = mndr.findCycles()
    print("Cycles = " + str(cycles))
    print("CyclesInd = " + str(cyclesInd))
    lengths = mndr.calcCycleLengths()
    print("Cycle lengths = " + str(lengths))
    maxL = max(lengths)
    i = lengths.index(maxL)
    print("The length of the largest cycle is " + str(maxL))
    mndr.drawCycle(cycles[i], ax = ax)
    
    drawAsPolygon(mndr)
    drawAsPolygon(mndr.reflected())
    
    plt.show()


if __name__ == '__main__':
    main()          
        
        


'''
A package to generate, manipulate and display meander systems.
It also handles some incidental objects such as permutations,
Dyck paths, and (non-crossing) pairings
'''

'''
#I prefer blank __init__.py for modules

from .Combinatorics import cycleLengths
from .Combinatorics import cyclesOfPerm
from .Combinatorics import enumPropMeanders
from .Combinatorics import rotationNumber
from .Combinatorics import random231perm

from .DyckPaths import randomDyckPath
from .DyckPaths import plotDyckPath
from .DyckPaths import areaDyckPath
from .DyckPaths import valleys
from .DyckPaths import peaks
from .DyckPaths import pathTo231perm

#from .Meander import Meander
from .Meander import areaMeander
from .Meander import randomMeander
from .Meander import randomComb
from .Meander import randomCrossSemiMeander
from .Meander import rainbowMeander
from .Meander import smallRainbowsMeander
from .Meander import drawAsPolygon

from .Pairing import randomPairing
from .Pairing import randomCrossPairing
from .Pairing import addCrossing
from .Pairing import standardPairing
from .Pairing import rainbowPairing
from .Pairing import rainbows
from .Pairing import smallRainbows
from .Pairing import prng2path

from .Utility import genPairings
from .Utility import allPairings
from .Utility import allMeanders
from .Utility import dot
from .Utility import dotMatrix
from .Utility import randomTree
from .Utility import randomCrossMeander
from .Utility import treeToDyckPath
from .Utility import plotRandomProperMeander
from .Utility import makeMeanderFromCycle
'''





'''
Created on Jul 21, 2020

@author: vladislavkargin
'''
import numpy as np
import random
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
#import matplotlib.lines as mlines


import rgs.mndrpy.Utility as ut
import rgs.mndrpy.DyckPaths as dp
#import mndrpy.ElementTL as tl


class Pairing(object):
    '''
    classdocs
    
    Non-crossing pairing on 2n point. Represented by a tuple prng,
    which is something like (3, 2, 1, 0) meaning that 0 is connected to 3
    and 1 is connected to 2. The tuple (1, 0, 3, 2) means that 0 connected to 1
    and 2 connected to 3. 
    
    Can also handle pairings with crossings. In this case a crossing matrix can be 
    calculated. 
    '''


    def __init__(self, path = None, prngArray = None):
        '''
        Constructor
        creates a pairing from a Dyck path (which is a sequence of 1 and -1), 
        or from an external pairing -- which is an array of numbers. 
        For example [3 2 1 0] pairs  0 with 3 and 1 with 2 
        '''
        if prngArray != None:
            self.prng = tuple(prngArray)
            return
        
        path = list(path)
        self.prng = []
        if path != None:
            stack = []
            self.prng = [0] * len(path)
            for i in range(len(path)):
                if path[i] == 1:
                    stack.append(i)
                else:
                    j = stack.pop(-1);
                    self.prng[i] = j;
                    self.prng[j] = i;    
        self.prng = tuple(self.prng)
        #self.crossM = None #if the matrix is crossing this should be calculated
        #separately
        
    def __str__(self):
        return str(self.prng)
    
    def __eq__(self, other):
        '''Overrides the default implementation'''
        if isinstance(other, Pairing):
            if self.prng == other.prng: 
                return True
            else: 
                return False
        return False
    
    def __hash__(self):
        return hash(tuple(self.prng))
    
    def size(self):
        '''return n for a pairing on 2n objects '''
        return len(self.prng)//2
    
    def draw(self, ax = None, up = True, asPath = False):
        ''' draws the pairing. 
        If up is False, then draws it in the lower half-plane.
        If asPath is true, draws it as the corresponding Dyck path.
        '''
        
        if ax == None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
        
        if not asPath:
            width = len(self.prng)
            ax.set_xlim(0, width)
            ax.set_ylim(-width/2, width/2)
            
            if up:
                theta1 = 0.0
                theta2 = 180.0
            else:
                theta1 = 180.0
                theta2 = 360.0
            for i in range(width):
                j = self.prng[i]
                xy = (i + (j - i)/2, 0)
                w = np.abs((j - i)/2)
                #print("i = " + str(i) + " j = " + str(j) + " xy = " + str(xy) 
                #      + " w = " + str(w))
                arc = mpatches.Arc(xy, 2*w, 2*w, 0, theta1, theta2)
                ax.add_patch(arc)
            #ax.grid(True)
            ax.xaxis.set_ticks(np.arange(0,width))
            ax.yaxis.set_ticks(np.arange(-width/2 + 1,width/2 , width/2 - 1))
        else: 
            path = prng2path(self)
            if up:
                dp.plotDyckPath(path, ax)
            else:
                dp.plotDyckPath(path, ax, method = "lowerLatticePath")
        pass

def randomPairing(n, w = None, seed = None):
    '''generates a (uniformly) random non-crossing pairing of length 2n
    If w != None, then generates a pairing from a Galton-Watson tree
    with weight sequence w'''
    if w == None:
        path = dp.randomDyckPath(n, seed = seed)
    else:
        tree = ut.randomTree(n + 1, w, seed = seed)
        path = ut.treeToDyckPath(tree)
    prng = Pairing(path)
    return prng

def randomCrossPairing(n, seed = None):
    ''' generate a random pairing of length 2n,
     which is allowed to have crossings.'''
    if seed != None:
        np.random.seed(seed)
    elements = list(range(2*n))
    prngArray = [0] * (2 * n)
    for k in range(0, 2*n, 2):
        ind1 = np.random.randint(2 * n - k)
        a = elements.pop(ind1)
        ind2 = np.random.randint(2 * n - k - 1)
        #print("ind2 = ",ind2)
        #print("elements = ", elements)
        b = elements.pop(ind2)
        prngArray[a] = b
        prngArray[b] = a
    return Pairing(prngArray = prngArray)


#def calcCrossM(self):
#    ''' calculate crossing matrix '''
#    #TODO
#    N = len(self.prng)
#    self.crossM = np.zeros((N,N))

def addCrossing(prng, i, j):
    '''
    Add a crossing: if we have pairs (i l) (j k) then changes this to (i k) (j l)
    [if i and j are paired, does nothing]
    retunrs a prng with the crossing
    '''
    if prng.prng[i] == j:
        return prng
    else: 
        l = prng.prng[i]
        k = prng.prng[j]
        prngArray = list(prng.prng)
        prngArray[i] = k 
        prngArray[k] = i
        prngArray[j] = l
        prngArray[l] = j 
        return Pairing(prngArray = prngArray)
        
def standardPairing(n):
    '''
    creates the pairing of half-length n 
    that connects 0 to 1, 2 to 3, 4 to 5 and so on 
    '''
    p = Pairing()
    p.prng = [-1] * 2 * n
    for i in range(n):
        p.prng[2 * i] =  2 * i + 1
        p.prng[2 * i + 1] = 2 * i
    return p

def rainbowPairing(n):
    '''
    creates the rainbow that pairs 0 to 2n - 1, 1 to 2n-2 and so on
    '''
    p = Pairing()
    p.prng = [-1] * 2 * n
    for i in range(n):
        p.prng[i] = 2 * n - 1 - i
        p.prng[2 * n - 1 - i] = i
    return p


    

def rainbows(widths):
    '''
     * Creates a NC pairing on 2*n points that consists of rainbows.
     * The half-widths of rainbows are contained in widths array.
     * 
     * For example the array [n] should correspond to the identity pairing
     * and the array [1, 1, ..., 1] - to the standard pairing. 
     * 
     * @param widths array of half-widths of rainbows
     * @return a pairing
     '''
    N = sum(widths)
    prng = [0] * (2*N)
    pos = 0
    for j in range(len(widths)):
        w = widths[j];
        for i in range(w):
            prng[pos + i] = pos + (2 * w - 1 - i)
            prng[pos + (2 * w - 1 - i)] = pos + i
        pos = pos + 2 * w;
    return Pairing(prngArray = prng)

def smallRainbows(amounts, sizes):
    '''
    creates a random pairing whose arks are small rainbows of two types
    parameters:
    amounts: number of rainbows of every type 
    sizes: half-lengths of the types
    for example amounts = [3, 4] sizes = [2, 3]
    means that we will rhave 3 rainbows of half size 2 and 4 rainbows of 
    half-size 3
    '''
    widths = ([sizes[0]] * amounts[0]) + ([sizes[1]] * amounts[1])
    random.shuffle(widths)
    prng = rainbows(widths)
    return prng

def prng2path(prng):
    '''
     * Calculates a Dyck path from a NC pairing
     * 
     * 
     * @param pairing an NC pairing 
     * @return corresponding Dyck path [0 are up, 1 are down] [1 - up, -1 down]
     '''
    N = len(prng.prng)
    path = np.ones(N)

    for i in range(N):
        if prng.prng[i] < i :
            path[i] = -1
    return path.astype(int);

def prng2prtn(prng):
    '''
    Calculate a NC partition on n elements from an NC pairing on 2n elements
    '''
    prtn = []
    N = len(prng.prng)
    marked = [False] * N #those elements that have been already processed.
    #initialization
    for x in range(0,N,2):
        if marked[x]:
            continue
        piece = []
        counter = 0
        while counter < 10000:
            if counter == 9999:
                print("Warning from prng2partition: the cycle was repeated 10,000 times")
            piece.append(x)
            marked[x] = True
            y = prng.prng[x]
            piece.append(y)
            x = y - 1
            if marked[x]:
                break
        prtn.append(piece)
    #now reduce it by halving
    prtn1 = []
    for piece in prtn:
        piece1 = []
        for x in piece:
            if x % 2 == 0:
                piece1.append(x//2)
        prtn1.append(piece1)
                   
    return prtn1
                
        
    
     

'''
For testing methods
'''
def main():
    
    n = 10
    seed = None    
    prng = randomPairing(n, seed)
    #prng = standardPairing(n)
    #prng = rainbowPairing(n)
    #prng = randomPairing(n, w = [1, 1, 1])
    #prng = randomCrossPairing(n)
    #prng = smallRainbows([2,3],[1,4])
    print(prng.prng)
    print("Size is ", prng.size())
    prng.draw()
    
    path = prng2path(prng)
    print(path)
    dp.plotDyckPath(path)
    
    prtn = prng2prtn(prng)
    print(prtn)
    
   
    #newPrng = addCrossing(prng, 0, 1)
    #newPrng.draw()
    '''
    prng = Pairing(path = None, prngArray = [2, 3, 0, 1])
    prng.draw(up = False)
    prng2 = Pairing(path = None, prngArray = [1, 0, 3, 2])
    print(prng == prng2)
    print(hash(prng))
    '''
    

    
    plt.show()

if __name__ == '__main__':
    main()    
    


'''
Created on Jan 24, 2020

@author: vladislavkargin
'''
import numpy as np
import matplotlib.pyplot as plt

import rgs.pmaps.TreeUtility as tu
from rgs.pmaps.PlanarGraph import PlanarGraph

import rgs.mndrpy.DyckPaths as dp



class OrderedTree(object):
    '''
    This is an object that represent a planar rooted tree.

    It is essentially a tree graph with additional structure
    and some convention on what the order of adjacency lists means
    In particular, 
    (0) Usually the root is 0. 
    (1) we will use the convention that the first element in the
    list of adjacent vertices represent the parent with exception 
    of the root vertex. Howeve, if the root is set with setRoot to something != 0,
     then this convention is not ensured. If one wants to change the root and ensure this convention
      then one should run the dfsRename method
    (2) the order of the elements in the adjacency list should correspond to the order in the 
    planar embedding of the tree, in counterclockwise order starting from the edge to
     the parent vertex. 
    '''


    def __init__(self, V = 1, root = 0):
        self.graph = PlanarGraph(V, root)
        self.root = root #this is the root 
                        #[it looks redundant to me. the info about the 
                        #root is already in the graph structure.
        #self.edgeLabels = np.array([])  #This is only used for some labeled trees 
                                        #[so far never used]
        #self.graph.labels = [] #[I think it is never used]
        

    def __str__(self):
        '''
        Returns a string representation of this tree.
        return the graph
        '''
        s = 'Graph of Tree: \n' + str(self.graph) + '\n'
        return s   
    
    def __repr__(self):
        '''
        Returns a string representation of this tree.
        return the graph
        '''
        s = str(self.graph)
        return s     
    
    def setRoot(self, root):
        '''
        changes root of the tree
        the deficiency of this method is that it does not ensure the 
        assumption that the first element in the adjacency list of each vertex 
        except root
        is the parent of the vertex
        '''
        self.root = root
        self.graph.root = root
    
    
    def outDegree(self, v):
        d = len(self.graph.adj[v])
        if v == self.root:
            outdegree = d
        else:
            outdegree = d - 1
        return outdegree
    

    def leaves(self):
        '''
    Returns all the leaves 
        '''
        lvs = []
        for v in range(self.graph.V):
            if self.outDegree(v) == 0:
                lvs.append(v)
        return lvs 
    
    def total_height(self):
        ''' returns the total height of the tree ''' 
        heights, _ = self.calcHeightsSizes(0)
        return sum(heights)
    
    
    def height(self):
        '''
        returns the height of the tree
        '''
        leaves = self.leaves()
        H = 0
        V = 0
        for v in leaves:
            h = self.heightVertex(v)
            if h > H:
                H = h
                V = v
        return H, V
    
    def heightVertex(self, v):
        '''
        measures the distance of vertex v from the root
        '''
        if v == self.graph.root:
            return 0
        dist = 0
        while v != self.graph.root:
            dist = dist + 1
            v = self.graph.adj[v][0] 
        return dist    
    
    def sizeVertex(self, v):
        '''
        calculate the size of the tree below the vertex (this vertex should 
        be different from the root)
        '''
        stack = [v]
        A = [-1]* self.graph.V
        A[v] = 0
        n = 1
        while (len(stack) > 0):
            #print("stack = " + str(stack));
            v = stack[-1]; #peaking at the top of the stack
            #print("processing v = " + str(v))
            flag = False #can we find unprocessed neighbors?
            d = self.graph.degree(v) - 1 #(outdegree)
            for i in range(d):
                w = self.graph.adj[v][i + 1]
                if (A[w] < 0): #this neighbor has not been explored previously
                    A[w] = n
                    n = n + 1
                    stack.append(w)
                    flag = True
                    break #stop searching for an uexplored neighbor
                else: 
                    continue
            if (not flag): 
                stack.pop() #no unexplored neighbor around v. Remove it from consideration.
        return  n  

    
    
        
    def calcHeightsSizes(self, v):    
        '''
    Traverses the tree using dfs starting with v and 
    returns two arrays. The first has the height of each vertex x, which is its distance from
    the root, the second is the size of each vertex x, which is the number of vertices in the 
    subtree rooted at x ( x itself is counted).
        '''
        
        #A[x] is the order in which x appears in the dfs path
        A = [-1]* self.graph.V
        heights = [-1]* self.graph.V 
        sizes = [-1] * self.graph.V
        stack = [v]
        A[v] = 0
        
        heights[v] = 0
        #sizes[v] = self.graph.V - 1 #in principle, this is redundant

        n = 1
        while (len(stack) > 0):
            #print("stack = " + str(stack));
            v = stack[-1]; #peaking at the top of the stack
            #print("processing v = " + str(v))
            flag = False #can we find unprocessed neighbors?
            d = self.graph.degree(v)
            for i in range(d):
                w = self.graph.adj[v][i]
                if (A[w] < 0): #this neighbor has not been explored previously
                    A[w] = n
                    n = n + 1
                    stack.append(w)
                    heights[w] = heights[v] + 1
                    flag = True
                    break #stop searching for an uexplored neighbor
                else: 
                    continue
            if (not flag): 
                stack.pop() #no unexplored neighbor around v. Remove it from consideration.
                sizes[v] = n - A[v]
                
        return  [heights, sizes]  
    '''
    TODO This is not yet implemented
    We need to to some re-ordering of vertices in the adjacency list of the 
    graph
    '''
    '''
    def setRoot(self, root = 0):
        self.root = root
        self.graph.setRoot(root)
    '''
    

    def dfsGeneral(self, v):    
        '''
    A general variant of dfs suitable for unrooted graph. It is starts with v and 
    returns the pair A, dfsPath:
    A[x] is the order in which x appears in the dfs path 
    dfsPath is a sequence of edges in the order in which they were
    first traversed (does not include returns through these edges)
        '''
        A = [-1]* self.graph.V #may be I do not need this 
        stack = [v]
        A[v] = 0
        dfsPath = []
        n = 1
        while (len(stack) > 0):
            #print("stack = " + str(stack));
            v = stack[-1]; #peaking at the top of the stack
            #print("processing v = " + str(v))
            flag = False #can we find unprocessed neighbors?
            d = self.graph.degree(v)
            for i in range(d):
                w = self.graph.adj[v][i]
                if (A[w] < 0): #this neighbor has not been explored previously
                    A[w] = n
                    n = n + 1
                    stack.append(w)
                    #print(dfsPath)
                    dfsPath.append([v, w])
                    flag = True
                    break #stop searching for an uexplored neighbor
                else: 
                    continue
            if (not flag): 
                stack.pop() #no unexplored neighbor around v. Remove it from consideration.
        return  [A, dfsPath]      
    

    def dfsRename(self, v):
        '''
    Rename all vertices of the tree in dfs order starting with vertex v.
    We will do it by creating a new graph
        '''
        [A, dfsPath] = self.dfsGeneral(v)
        V = self.graph.V
        gr = PlanarGraph(V)
        for i in range(V - 1):
            edge = dfsPath[i]
            gr.addEdge(A[edge[0]],A[edge[1]])
        self.root = 0
        self.graph = gr
        self.graph.root = 0
            
    

    def draw(self, drawLabels = False, rootInCenter = False, block = False, 
             calculateRepr = "No", ax = None, rotate = False):
        '''
        Draws the tree using graph.draw function. 
            Parameters:
            - drawLabels : if True, draws the labels, 
            - rootInCenter: if True, tries to put the root node in center 
            - block: if block is true, then calling 
                the draw function will stop execution of further commands. 
                if block is false (default), then the execution continues 
                but the calling function
                should have plt.show() at the end. 
            - calculateRepr -- passes this to the graph.draw, however the 
                default representation puts the leaves on the circle and calculate the rubber
                representation for other vertices. 
            - ax if given, draws the tree in the specified axis. 
            
            Returns:
             fig and ax handles
                
        '''
        fixiki = self.leaves() #if we want a correct graphical representation
                            #the leaves should be arranged in a correct order
                            #the easiest way to do it is to 
                            #rename vertices in DFS order (dfsRename method)
        if (rootInCenter): #put root in center: can violate planarity (non self-intersection)
            self.graph.reprOnCircle(fixiki)
            fixiki.append(self.root)
            self.graph.reprVertices([self.root],[[0, 0]])
        else: #treat root as a usual leaf:
            fixiki.append(self.root)
            self.graph.reprOnCircle(fixiki)
        self.graph.calcRubberRepr(fixiki)
        if rotate:
            for count, point in enumerate(self.graph.repr):
                self.graph.repr[count][0] = - point[0] 
                self.graph.repr[count][1] = - point[1] 
        #print(self.graph.repr)
        fig1, ax1 = self.graph.draw(root = 0, calculateRepr = calculateRepr,
                drawLabels = drawLabels, block = block, ax = ax)
        return fig1, ax1
    
    def drawPathToVertex(self, ax, v):
        '''
        draws a path that connects a vertex v to the root in the axes given by the argument ax
        The axes should be created by the draw() function
        ''' 
        path = [v];
        if v == self.graph.root:
            print("drawPathToVertex says: the vertex should be different from the root")
            return 
        else:
            while v != self.graph.root:
                v = self.graph.adj[v][0]
                path.append(v)
            self.graph.drawPath(ax, path)
         
    
    #def setLabels(self, labels):
        '''
        set the labels of vertices
        '''
    #    self.graph.labels = labels
        
def fromGraph(gr):
    '''
    builds the tree from the graph gr (which should be a tree topologically). 
    The vertices are renamed in dfs order.  
    '''      
    tr = OrderedTree(V = gr.V, root = gr.root)
    tr.graph = gr
    tr.dfsRename(gr.root)
    return tr
    

def fromLukasiewicz(path):
    '''
    Returns a rooted planar tree created from a Lukasiewicz path
    '''
    V = len(path)
    #print(path)
    tr = OrderedTree(V)
    stack = [0] #put root on the stack
    x = 1 #current available vertex
    for i in range(V):
        v = stack.pop()
        #print("working on " + str(v))
        k = int(path[i] + 1)
        if (k > 0):
            #print("k =" + str(k));
            for _ in range(k):
                tr.graph.addEdge(x, v)
                stack.append(x)
                x = x + 1
        #print(stack)
    tr.dfsRename(0)
    return tr


def fromDyck(path):
    '''Bilds a planar tree from a Dyck path of 1 and - 1'''
    V = len(path)//2 + 1
    tr = OrderedTree(V)
    unmarked = list(range(V))
    stack = [0] #put root on stack
    unmarked.remove(0);
    for i in range(len(path)):
        #print("stack = ", stack)
        #print("unmarked = ", unmarked)
        if path[i] == 1:
            v = unmarked[0]
            unmarked.remove(v)
            tr.graph.addEdge(stack[-1], v)
            stack.append(v)
        else:
            stack.pop()
    return tr        
            
    

def randomTree(n, w, SEED = 0):
    '''returns a random simply generated tree on n vertices using the weights in w
    The algorithm works by building a random Lukasiewicz path and converting it to tree
    '''
    path = tu.randomLukasiewicz(n, w, SEED = SEED)
    tr = fromLukasiewicz(path) 
    return tr  


def randomBTree(n, SEED = 0):
    '''
    returns a random B tree with n inner nodes each of which has 2 leaves
    This is used for building a triangulation
    '''
    path = tu.randomLukas31(n, SEED = SEED)
    tr= tu.decodeToPSTree(path)  
    return tr

def randomSlimTree(k, n, W = [], SEED = 0):
    '''
    generates a random Galton-Watson tree that has k leaves and n vertices
    Arguments:
    k - number of leaves
    n - number of vertices 
    W - if it is empty then the function assumes that the tree is binary/unary,
        if not we will use it as the weight sequence for the original Galton-Watson tree
    returns the random tree 
    
    TODO - improve to be able to work for any distribution of weights
    '''
    if len(W) == 0:
        path = tu.randomLukasBinary(k, n, SEED = SEED)
        tree = fromLukasiewicz(path)
    else:
        w1 = tu.rescale(W)
        #print(w1)
        path = tu.randomLukasSlim(k, n, w1, SEED = SEED)
        #print("path = ", path)
        tree = fromLukasiewicz(path)
        #print(tree)
    return tree




'''
creates a random embedding of the tree by 
assigning labels from {-1, 0, 1} to edges
'''
'''
def randomEdgeLabel(self):
    self.edgeLabels=np.random.randint(-1, 2, self.V - 1)
'''

'''  
visits every vertex by DFS and calculates the vertex labels
'''    
'''    
def calcVertexLabels(self):
    stack = [self.root]
    path = []
    self.vertexLabels[self.root] = 1;
    edge = 0;
    while (len(stack) > 0):
        x = stack[-1] #current element on top of the stack
        path.append(x)
        if (self.outdegree(x) == 0): #this is a leaf, do not go deeper
            stack.pop()
        else:
            children = self.graph.adj[x].remove(0)
            for child in children:
                self.vertexLabels[child] = self.vertexLabels[x] + self.edgeLabels[edge]
                edge = edge + 1
                stack.append(child)                
    pass
'''    
    

'''
For testing methods
'''
def main():
    print('OrderedTree methods are used')  
    w = np.array([1] * 3)
    #rw = tu.rescale(w) #This rescaling is done in function that generates Lukasiewicz path
    #print(rw)
    
    n = 10
    #seed = 12345
    seed = 0 #(random - changing from one execution to another)

    tr = randomTree(n, w, SEED = seed)
    #tr = randomBTree(n, SEED = seed) 
    print(tr)
    tr.draw(drawLabels = True)

    heights, sizes = tr.calcHeightsSizes(0)
    print("heights = " + str(heights))
    print("total_height = " + str(tr.total_height()))
    print("sizes = " + str(sizes))
    
    
    '''
    These are some drawings and calculations associated with Slim Trees
    '''
    
    '''
    n = 400
    #n = 20
    seed = 12345
    seed = 0
    k = 10
    tr = randomSlimTree(k, n, SEED = seed)
    #print(tr)
    #fig, ax = tr.draw(drawLabels = True, block = False)
    #fig, ax = tr.draw(drawLabels = False, block = False)  
    #fig1, (ax1, ax2) = plt.subplots(nrows=1, ncols=2)
    _, ax = tr.draw(drawLabels = False, block = False)  
    H, v = tr.height()
    dist = tr.heightVertex(v);
    print("Height of vertex " + str(v) + " is " + str(dist))
    tr.drawPathToVertex(ax, v)
    print("Height of the tree is " + str(H))  
    v = 1
    S = tr.sizeVertex(v)
    print("size of vertex " + str(v) + " is " + str(S))
    
    
    n = 5
    seed = 3
    path = dp.randomDyckPath(n, seed = seed)
    print("path = ", path)
    tr = fromDyck(path)
    
    _, ax = tr.draw(drawLabels = True, block = False)  
    '''
    plt.show() 
    
    
   
    
    
    
if __name__ == "__main__":
    main()
    


'''
Created on Jul 11, 2020

A class encapsulating a triangulation and associated functions.

@author: vladislavkargin
'''

import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
#from matplotlib import cm

import copy
#import math
import numpy as np

import rgs.pmaps.OrderedTree as ot
import rgs.pmaps.TreeUtility as tu
import rgs.pmaps.PlanarGraph as pg


class Triangulation(object):
    '''
    An object that represent a triangulation. It is a specialization of the planar graph
    so much of its functionality is provided by the PlanarGraph object.
    '''


    def __init__(self, tree = None):
        '''
        If tree is available buils the triangulation from the tree 
        using a variant of the Poulalhon-Scheffer method, 
        else builds a trivial triangulation that consists of 3 vertices.
        '''
        if tree != None:
            self.graph = pg.buildTriangulation(tree)
            self.realizer = self.calcRealizer()
        else:
            self.graph = pg.PlanarGraph(V = 3, root = 0)
            self.graph.rootArrow = 1
            self.graph.addEdge(0,1)
            self.graph.addEdge(1,2)
            self.graph.addEdge(0,2)
            self.realizer = ([-1, -1, -1], [-1, -1, -1], [-1, -1, -1]) 
    
    def __str__(self):
        return str(self.graph)
    
            
    def __eq__(self, other):
        """Overrides the default implementation. Triangulations are
        the same if their realizers are the same."""
        if isinstance(other, Triangulation):
            if self.realizer == other.realizer: #realizers are same
                return True
            else: 
                return False
        return False
    
    def __hash__(self):
        return hash(str(self.realizer))
    
    

    def canonLabelling(self):
        '''
        build a canonical labelling for a maximal planar graph (that is, for a triangulation)
        with outer face a, b, c
        That is, outputs a sequence [v0, v1, ..., v_n], where v0 = b, v1 = c, v_n = a, and the 
        sequence satisfies certain requirements useful for building a planar embedding of the
        graph. See Schnyder paper (1990) about realizers and references within. 
        It is assumed that the triangulation has at least 4 vertices  
        '''
        #TODO: I should enhance this by renaming so that the root was 0, arrowRoot 1 and 
        #the remaining vertex on the boundary 2. Perhaps it should be done in a separate 
        #method.
        gr = copy.deepcopy(self.graph)
        V = gr.V
        seq = [0] * V
          
        boundary = gr.getOuterBoundary(u0 = gr.rootArrow) #this sometimes fails I don't know
                                                        #why
        if boundary == -1:
            return -1                                              
        b = boundary[0]
        c = boundary[1]
        a = boundary[2]
        seq[0] = b 
        seq[1] = c
        seq[-1] = a
        
        for i in range(V- 3): #we are going to remove V - 3 non-boundary vertices
            nbrs = gr.adj[a]
            #print("neighbors are " + str(nbrs))
            #we are looking for a vertex v among neighbors of vertex a, 
            #which is not b or c and which 
            #has exactly 2 common neighbors with vertex a
            #In terminology of Schnyder, av is contractible.
            isContractionFound = False
            for v in nbrs:
                if (v == b) or (v == c):
                    continue
                else:
                    count = 0 
                    for u in gr.adj[v]: 
                        if (u in nbrs):
                            count = count + 1
                    if count == 2: #a contractible edge is found
                        isContractionFound = True
                        #print("Contractible edge is (" + str(a) + ", " + str(v) + ")")
                        break 
            if not isContractionFound:
                print("canonLabelling says: Contractible edge is not found")
                print("drawing current graph:")
                gr.draw(drawLabels = True)
                break
            else:
                seq[gr.V - 2 - i] = v
                #contracting the graph
                for u in gr.adj[v]: #first we add edges to a
                    #print(u)
                    u0 = v #this is to keep track of which new edges were added to a
                        #if no edges are added, this is v
                    if (u not in nbrs) and (u != a):
                        #print("Nontrivial vertex u = " + str(u))
                        gr.addEdgeBeforeAfter(u, a, v, u0)
                        u0 = u
                toRemove = copy.deepcopy(gr.adj[v])
                for u in toRemove: #now we remove all edges from v
                    #print("removing " + str(u))
                    gr.removeEdge(v, u)
                #print("After contraction of (" + str(a) + ", " + str(v) + ")")
                #print("The graph is " + str(gr))
                #gr.draw(drawLabels = True)
        return seq 
    
    def calcRealizer(self):
        '''
        Builds the Schnyder realizer of the triangulation graph. 
        Arguments:
        graph - a triangulation
        seq - a canonical labelling build by canonLabelling function
        Output:
        a triple of arrays, each of them gives a map from a vertex of a triangulation to its parent
        in one of 3 trees.
        if we have b root, c rootarrow, and a the remaining vertex on the outer triangle
        then T1 has its root at a, T2 - at b, and T3 - at c.
        [The array representation of trees loses some info about planar embeddings of these trees]
        '''
        return self.graph.calcRealizer()
        
            
    def draw(self, title = '', drawLabels = False, block = False, root = -1, 
             calculateRepr = "Schnyder", boundary = [], showDual = False,
             palette = "prism"):        
        fig, ax = self.graph.draw(title = title, drawLabels = drawLabels, block = block, root = root, 
             calculateRepr = calculateRepr, boundary = boundary, showDual = showDual,
             palette = palette)   
        return fig, ax
    
    def drawRealizer(self, ax, realizer):
        '''
        draws graph with a Schnyder's realizer (which was computed by calcRealizer)
        '''
        #fig, ax = self.draw(calculateRepr = calculateRepr)
        A = np.array(self.graph.repr)
        x = A[:, 0]
        y = A[:, 1]
        for i in range(3):
            if i == 0:
                c = "red"
            elif i == 1:
                c = "green"
            else:
                c = "purple"
            T = realizer[i]
            for i, v in enumerate(T):
                if v >= 0:
                    line = mlines.Line2D(np.vstack((x[i], x[v])),
                                     np.vstack((y[i], y[v])), marker = ".", color = c);
                    ax.add_line(line)
                    arrow = mpatches.Arrow(x[i], y[i], x[v] - x[i], y[v] - y[i], width = 0.1,
                                            color = c);
                    ax.add_patch(arrow)

def randomTriangulation(n, SEED = 0, method = "Lukas31"):
    '''
    builds a random triangulation on n vertices
    arguments:
    n - number of vertices in the triangulation
    SEED is the seed of the random generator for replication;
        if SEED = 0 (default) then the results will change from time to time.
    method - if "Lukas31" (default) uses the Poulalhon - Schaeffer method
        to build a uniformly random triangulation based on building a Lukasiewisz path
        with steps 3 and -1 and then decoding it to a tree, in which each internal vertex
        is connected to two leaves.
        -if "decoratedPath" then modifies the Poulalhon Schaeffer method and builds 
        a tree in which an internal vertex is connected to two leaves and exactly one next
        internal vertex, so it is a path decorated with leaves (or some kind of asymmetric
        caterpillar. 
    returns the triangulation and the tree from which it was constructed
    '''
    if method == "Lukas31":
        tree = ot.randomBTree(n - 2, SEED = SEED)
        #tree.draw(drawLabels = True)
    elif method == "decoratedPath":
        tree = tu.decoratedPathTree(n, SEED = SEED)
    else:
        print("randomTriangulation says: the method should be either Lukas31 or decoratedPath")
    trngl = Triangulation(tree)
    return trngl, tree 


'''
For testing methods
'''
def main():
    print('Triangulation.main() started')
    
    

    
    #n = 8
    #seed = 23167
    '''
    n = 5
    seed = 0
    trngl, _ = randomTriangulation(n, SEED = seed)
    #trngl = randomTriangulation(n, SEED = seed, method = 'decoratedPath')
    print(trngl.graph)
    _, ax = trngl.draw(drawLabels = True)
    

    profile = trngl.graph.degreeProfile()
    print("Degree profile is " + str(profile))
    print("Max degree = " + str(trngl.graph.maxDegree()))
    boundary = trngl.graph.getOuterBoundary(u0 = trngl.graph.rootArrow)
    print("Calculated Boundary = " + str(boundary))
    seq = trngl.graph.canonLabelling()
    print("Canonical Labeling = " + str(seq))
    
   # realizer = trngl.graph.calcRealizer()
    #print(trngl.realizer) 
    
    X = trngl.graph.calcSchnyderRepr()
    print("barycentric coordinates are: \n" + str(X))   
    
    #_, ax = trngl.draw(drawLabels = True)
    #trngl.drawRealizer(ax, trngl.realizer)
    '''
    
    
    n = 5
    seed = 0
    hashTrngl = {}
    ITER = 100
    for count in range(ITER):
        if count % 10 == 0:
            print(count)
        trngl, _ = randomTriangulation(n, SEED = seed)
        #print(str(trngl))
        seq = trngl.canonLabelling()
        if seq == -1:
            break
        if trngl in hashTrngl.keys():
            hashTrngl[trngl] = hashTrngl[trngl] + 1
        else:
            hashTrngl[trngl] = 1
    print(len(hashTrngl))
    for i in range(len(hashTrngl)):
        trngl = list(hashTrngl.keys())[i]
        _, ax = trngl.draw(drawLabels = True)
        trngl.drawRealizer(ax, trngl.realizer)
        

    
    plt.show()  
    
    print('Triangulation.main() finished')  

    
if __name__ == "__main__":
    main()


'''
Created on Mar 12, 2021

@author: vladislavkargin

A collection of functions for Dyck paths.
A Dyck path is an array of +1 and -1, such that the total sum is zero and 
the partial sums are always non-negative. 
'''

import rgs.mndrpy.Pairing as pg
import rgs.mndrpy.Utility as ut

import numpy as np
import matplotlib.pyplot as plt
#from numpy.random import MT19937
#from numpy.random import RandomState, SeedSequence


def randomDyckPath(n, seed = None):
    '''
     * Generates a random Dyck path with 2n steps
     * 
     * @param n half-size of path
     * @return a sequence of -1 and 1 steps.
    '''
    
    if seed != None:
        #rs = RandomState(MT19937(SeedSequence(seed)))
        np.random.seed(seed)
    else:
        np.random.seed()    
        
        
    path = [1] * n + [-1] * (n + 1)   
    #shuffle randomly
    np.random.shuffle(path)
    
    #do a cyclic shift. 
    #Find the first occurrence of the minimum
    S = np.cumsum(np.array([0] + path))
    '''
    fig1, ax1 = plt.subplots(nrows=1, ncols=1)
    ax1.plot(S)
    ax1.grid(True)
    '''
    m = np.min(S)
    positions = np.where(S == m)[0]
    pos = positions[0]
    #print(pos)
    path1 = path[pos:] + path[:pos]
    del path1[-1]
    return path1

def allDyckPaths(n):   
    '''returns a list of all pairings of length 2n'''     
    A = []
    for pairing in ut.allPairings(n):
        path = pg.prng2path(pairing)
        #print(path)
        A.append(path)
    return A


def plotDyckPath(path, ax = None, method = "upperLatticePath"):
    ''' plots a Dyck path. Path is either list of 1, -1 or numpy array
    of 1, - 1'''
    if isinstance(path, list):
        n = int(len(path)/2)
    else:
        n = int(path.shape[0]/2)
    X, Y = buildPlanePath(path, method = method)
    if ax == None:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
    major_ticks = np.arange(0, n + 1, 1)
    #minor_ticks = np.arange(0, 101, 5)
    ax.set_xticks(major_ticks)
    #ax.set_xticks(minor_ticks, minor=True)
    ax.set_yticks(major_ticks)
    #ax.set_yticks(minor_ticks, minor=True)
    # And a corresponding grid
    ax.grid(which='both')
    ax.plot(X, Y)
    Z = range(n + 1)
    ax.plot(Z)
    return ax  

    
def jeulin_transform(path):
    ''' jeulin transform of the path '''
    L = profile(path)
    H = inv_profile_cdf(path)
    T = []
    for x in H:
        T.append(L[x])
    return T


def inv_profile_cdf(path):
    '''
    "right-continuous" inverse of the profile cdf"
    ''' 
    if isinstance(path, list):
        n = int(len(path)/2)
    else:
        n = int(path.shape[0]/2)
    H = profile_cdf(path)
    H_inv = []
    for x in range (2 * n + 1):
        pos = np.where(H >= x)[0]
        H_inv.append(pos[0])
    return np.array(H_inv)
    

def profile_cdf(path):
    ''' the array [H(1), H(2), H(x) ] where H(x) is the sum of 
    local times l(t) for t <= x '''
    H = np.cumsum(profile(path))
    return H

def GS_statistic(path):
    ''' calculate the Gorin-Shkolnikov statistic '''
    if isinstance(path, list):
        n = int(len(path)/2)
    else:
        n = int(path.shape[0]/2)
    x = (area(path) - sum_squared_profile(path)/2)/(n ** (3/2))
    return x
    
def sum_squared_profile(path):    
    ''' calculate the sum of l(x) where l(x) is occupation time for x defined
    in profile(path)'''
    prfl = profile(path)
    S = np.sum(np.square(prfl))
    return S

def profile(path):
    '''
    calculate the occupation times for the Dyck path
    (for the excursion it would be local times l(x))
    The first zero is not counted.
    '''
    if isinstance(path, list):
        n = int(len(path)/2)
    else:
        n = int(path.shape[0]/2)
    S = np.cumsum(np.array(path))
    #print(S)
    prfl = []
    for x in range(n):
        pos = np.where(S == x)[0]
        l = np.size(pos)
        #print(l)
        if l > 0:
            prfl.append(l)
        else:
            break
    return np.array(prfl)    


def height(path):
    ''' calculate the height of the given Dyck path'''
    S = np.cumsum(np.array(path))
    return max(S)
    


def area(path):
    ''' calculate the area of the given Dyck path, defined as the number
    of unit squares under the path but above the line y = x'''
    if isinstance(path, list):
        n = int(len(path)/2)
    else:
        n = int(path.shape[0]/2)
    x = 0
    y = 0
    S = [(x, y)]
    for i in range(2 * n):
        if path[i] == 1:
            y += 1
        else:
            x += 1
        S.append((x,y))
    area = 0
    y_prev = 0
    for p in S: #every time as y increases, we add y - x - 1 to the area
        if p[1] > y_prev:
            area += p[1] - p[0]- 1
            y_prev = p[1]
    return area

def rank(path):
    ''' calculates the rank of the path, defined as the number of parts in 
    the corresponding non-crossing partition
    '''
    prtn = pg.prng2prtn(path2prng(path))
    rank = len(prtn)
    return rank

def valleys(path):
    '''return the list of valleys in the path'''
    if isinstance(path, list):
        n = len(path)//2 
    else:
        n = path.shape[0]//2
    valleys = []
    for i in range(2 * n - 1):
        if path[i] == - 1 and path[i + 1] == 1:
            valleys.append(i)
    return valleys

def peaks(path):
    '''find peaks of the given Dyck path,
    and returns their list'''
    if isinstance(path, list):
        n = len(path)//2 
    else:
        n = path.shape[0]//2
    peaks = []
    for i in range(2 * n - 1):
        if path[i] == 1 and path[i + 1] == -1:
            peaks.append(i)
    return peaks


def buildPlanePath(path, method = 'upperLatticePath'):
    ''' creates the path in the plane '''
    if isinstance(path, list):
        n = int(len(path)/2)
    else:
        n = int(path.shape[0]/2)
    x = 0
    y = 0
    S = [(x, y)]
    if method == "upperLatticePath":
        for i in range(2 * n):
            if path[i] == 1:
                y += 1
            else:
                x += 1
            S.append((x,y))
    else:         
        for i in range(2 * n):
            if path[i] == 1:
                x += 1
            else:
                y += 1
            S.append((x,y))
    #print(S)
    X, Y = list(zip(*S)) 

    return X, Y

def pathTo231perm(path):
    '''converts a Dyck path to 231 avoiding permutation
    I am using the description of the bijection from Peterson's "Eulerian paths"
    '''
    #first build the reflected path 
    X, Y = buildPlanePath(path, method = 'lowerLatticePath')
    n = (len(X) - 1)//2
    #One-by-one, for every column from right to left,
    #Find the lowest unoccupied row
    perm = [0] * n 
    used_rows = []
    for x in reversed(range(1, n + 1)):
        y = Y[X.index(x)] # this is the bound from below that the path gives us
        while y in used_rows:
            y += 1
        used_rows.append(y)
        perm[x - 1] = y
    return perm
    
def path2prng(path):
    '''converts a Dyck path to a non-crossing paring.
    '''
    prng = pg.Pairing(path = path)
    return prng


'''
For testing methods
'''
def main():
    
    n = 5
    seed = 3
    path = randomDyckPath(n, seed = seed)
    pks = peaks(path)
    print(path)
    print(pks)
    plotDyckPath(path, method = 'upperLatticePath')
    perm = pathTo231perm(path)
    print(perm)
    
    paths = allDyckPaths(n)
    print(len(paths))
    
    n = 400
    path = randomDyckPath(n)
    plotDyckPath(path)
    prng = path2prng(path)
    prng.draw()
    
    
    print("Height = ", height(path))
    print("Area = ", area(path))
    print("Rank = ", rank(path))
    print("Profile = ", profile(path))
    print("Profile_cdf = ", profile_cdf(path))
    print("Inverse_profile_cdf = ", inv_profile_cdf(path))
    T = jeulin_transform(path)
    #print("Jeulin = ", T)
    print("GS = ", GS_statistic(path))
    #It looks that GS is not centered. 
    plt.figure()
    plt.plot(T)
    
    
    plt.show()
    
    pass

if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rgs', 'rgs.mndrpy', 'rgs.pmaps', 'rgs.ribbons']

package_data = \
{'': ['*']}

install_requires = \
['itu.algs4', 'numdifftools', 'numpy']

setup_kwargs = {
    'name': 'rgs',
    'version': '0.1.11',
    'description': 'A small package to display and manipulate random geometry structures like pairings, meanders, triangulations and tilings.',
    'long_description': '# RandomGeometricStructures\nA collection of various function to draw and analyze random geometric structures (like pairings, tilings etc.)\n\nAn illustration can be found at: \n\nFor pmaps package:\n[Triangulations](https://colab.research.google.com/drive/1URQte3-X6Lau74B0JSjU7gkStO_-kx0O?usp=sharing )\n',
    'author': 'Vladislav Kargin',
    'author_email': 'slavakargin@yahoo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/slavakargin/RandomGeeometricStructures',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



'''
Created on Aug 14, 2020

@author: vladislavkargin
'''
import rgs.mndrpy.Meander as Meander
import rgs.mndrpy.Utility as ut
import rgs.mndrpy.Pairing as pg
import rgs.mndrpy.DyckPaths as dp


import numpy as np
import itertools 
import matplotlib.pyplot as plt
from sympy.combinatorics.permutations import Permutation
#from progress.bar import Bar


'''
I assume that permutation is coded as a list,
 obtained as permutation of range(n), 
'''


def inversions(perm):
    ''' 
    calculates the number of inversions for permutation perm.
    '''
    p = Permutation(perm)
    return p.inversions()
    

def get_cycles(perm):
    ''' calculates all cycles in the permutation perm. 
    For, example if the permutation is 102534, then return a list of 
    [01][2][354]. Currently, the order of cycles and the order inside cycles are NOT
    normalized to the standard order defined in Ennumerative Combinatorics by Stanley.
    '''
    cycles = []
    elements = list(range(len(perm)))
    while len(elements) > 0:
        cycle = []
        i = elements[0]
        elements.remove(i)
        cycle.append(i)
        j = perm[i]
        while j != i:
            elements.remove(j)
            cycle.append(j)
            j = perm[j]
        cycles.append(cycle)
    return cycles
        
def cycleLengths(perm):
    '''calсulates cycle lengths for a permutation. returns them in 
    decreasing order (as a partition of the length of the permutation'''
    cycles = get_cycles(perm)
    y = np.sort(list(map(len, cycles)))
    y = y[::-1]
    return y     

def enumPropMeanders(n):
    ''' enumerates proper Meanders on [2n]. (Obviously, n cannot be too large.)
    '''
    meanders = set([])
    for uPrng in ut.genPairings(n):
        for dPrng in ut.genPairings(n):
            mndr = Meander.Meander(uPrng, dPrng)
            if mndr.isProper():
                meanders.add(mndr)
    return meanders       

def enumIrrMeanders(n):
    ''' enumerates irreducible Meanders on [2n]. (Obviously, n cannot be too large.)
    NB. The definition that I use is different from the definition used by Zvonkin-Lando,
    in their paper about meanders,  
    so this is more of exploratory character.
    '''
    meanders = set([])
    for uPrng in ut.genPairings(n):
        for dPrng in ut.genPairings(n):
            mndr = Meander.Meander(uPrng, dPrng)
            if mndr.isIrreducible():
                meanders.add(mndr)
    return meanders     

def rotationNumber(cycle):
    '''calculate the rotation number for a cycle
    The input is a permuted list of numbers from 0 to 2n - 1 that starts with 0.
    It represents a cycle in the plane (a meander with self-intersections allowed)
    For example (0 2 1 3) represents a cycle formed by arcs (0 2) and (1 3) in 
    the upper-halfplane and arcs (2 1) (3 0) in the lower half-plane
    The function returns the rotation number for this cycle. For example, for this
    cycle the rotation number is 2. For non-crossing meanders the rotation number
    is always 0. '''
    if len(cycle)%2 != 0:
        print("Rotation Number says: The length of the cycle must be even!")
        return
    n = int(len(cycle)/2)
    #print(n)
    rn = 0
    for i in range(2*n):
        if cycle[(i + 1)%(2*n)] > cycle[i]:
            rn = rn + (-1)**(i%2)
        else:
            rn = rn - (-1)**(i%2)
    return int(rn/2) 
    

def random231perm(n):
    ''' generates a random 231 permutation of n objects, like 3 0 2 1 for n = 4'''
    path = dp.randomDyckPath(n)
    p = dp.pathTo231perm(path)
    return p   


def prtn2perm(prtn):
    '''
    calculates a permutation from a non-crossing partition.
    For example, if prtn = [[0], [1], [2, 9, 8, 7], [3, 5, 4], [6]]
    then it is converted to a permutation 
    [[2, 7, 8, 9], [3, 4, 5]] in cycle notation
    or 
    [0, 1, 7, 4, 5, 3, 6, 8, 9, 2] in the word notation. 
    '''
    N = 0
    for part in prtn:
        part.sort()
        m = max(part)
        if m > N:
            N = m
    
    perm = [0] * (N + 1)
    for part in prtn:
        for i in range(len(part) - 1):
            perm[part[i]] = part[i + 1]
        perm[part[len(part) - 1]] = part[0]
    return perm

    
'''
For testing methods
'''
def main():
    #n = 7
    n = 4
    meanders = enumPropMeanders(n)
    print(len(meanders))
    #print(list(map(str,meanders)))
    
    meanders = enumIrrMeanders(n)
    print(len(meanders))
    #print(list(map(str,meanders)))
    
    print(list(itertools.permutations(range(1,n))))
    cycle = [0, 1, 5, 2, 4, 3]
    mndr = ut.makeMeanderFromCycle(cycle)
    print(mndr)
    mndr.draw()
    rn = rotationNumber(cycle)
    print("Rotation number = ", rn) 
    
    cycle = [0, 5, 1, 4, 2, 3]
    mndr = ut.makeMeanderFromCycle(cycle)
    print(mndr)
    mndr.draw()
    rn = rotationNumber(cycle)
    print("Rotation number = ", rn) 
    
    
    #a random 231 permutation 
    n = 100
    p = random231perm(n)
    print(p)
    plt.figure()
    plt.plot(p,'*')
    plt.grid()

    cycles = get_cycles(p)
    print(cycles)
    print(cycleLengths(p))    
    
    #rng = np.random.default_rng()
    #perm = rng.permutation(10)
    perm = [7,3,2,6,4,5,1,0]
    perm = [1,2,5,4,3,7,6,0]
    print(perm)
    
    cycles = get_cycles(perm)
    print(cycles)
    print("inversions =", inversions(perm))
    
    prtn = [[0], [1], [2, 9, 8, 7], [3, 5, 4], [6]]
    perm = prtn2perm(prtn)
    print(perm)
    
    
    path = dp.randomDyckPath(n)
    dp.plotDyckPath(path)
    print("Area = ", dp.area(path))
    perm = prtn2perm(pg.prng2prtn(dp.path2prng(path)))
    print("Inversions = ", inversions(perm))
    
    plt.show()

if __name__ == '__main__':
    main()